/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.web.actions.utils;

import org.wso2.registry.*;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.session.UserRegistry;

import java.sql.Timestamp;
import java.util.*;

/**
 * Represents an advanced query on normal resources. Handles all details of defining, executing and
 * parameter processing of such queries.
 */
public class AdvancedResourceQuery {

    /**
     * List of parameters for the query. NOTE THAT THIS LIST SHOULD ALWAYS BE PROCESSED IN THE ORDER
     * THEY ARE LISTED BELOW.
     */
    private String resourceName;
    private String authorName;
    private String updaterName;
    private Date createdAfter;
    private Date createdBefore;
    private Date updatedAfter;
    private Date updatedBefore;
    private String commentWords;
    private String propertyName;
    private String propertyValue;

    private String tags;
    private String queryPath;
    private List<String> params = new ArrayList<String>();

    public Resource execute(Registry registry) throws RegistryException {

        // find if query is already defined
        // if defined, execute
        // else, create new query and execute

        String queryPath = computeQueryPath();
        if (!queryExists(queryPath)) {
            defineQuery(queryPath);
        }

        if (params.isEmpty()) {
            if (resourceName != null && resourceName.length() != 0) {
                params.add("%/" + resourceName);
            }

            if (authorName != null && authorName.length() != 0) {
                params.add(authorName);
            }

            if (updaterName != null && updaterName.length() != 0) {
                params.add(updaterName);
            }

            if (createdAfter != null) {
                params.add(new Timestamp(createdAfter.getTime()).toString());
            }

            if (createdBefore != null) {
                params.add(new Timestamp(createdBefore.getTime()).toString());
            }

            if (updatedAfter != null) {
                params.add(new Timestamp(updatedAfter.getTime()).toString());
            }

            if (updatedBefore != null) {
                params.add(new Timestamp(updatedBefore.getTime()).toString());
            }

            if (commentWords != null && commentWords.length() != 0) {
                params.add("%" + commentWords + "%");
            }

            if (tags != null && tags.length() > 0) {
                String[] tag = tags.split(" ");
                if (tag.length > 0) {
                    params.addAll(Arrays.asList(tag));
                }
            }

            if (propertyName != null) {
                params.add(propertyName);
            }

            if (propertyValue != null) {
                params.add(propertyValue);
            }
        }

        Map paramMap = new HashMap();
        for (int i = 0; i < params.size(); i++) {
            Object value = params.get(i);
            paramMap.put(Integer.toString(i + 1), value);
        }

        return registry.executeQuery(queryPath, paramMap);
    }

    public String getResourceName() {
        return resourceName;
    }

    public void setResourceName(String resourceName) {
        if ("".equals(resourceName)) resourceName = null;
        this.resourceName = resourceName;
    }

    public String getAuthorName() {
        return authorName;
    }

    public void setAuthorName(String authorName) {
        if ("".equals(authorName)) authorName = null;
        this.authorName = authorName;
    }

    public String getUpdaterName() {
        return updaterName;
    }

    public void setUpdaterName(String updaterName) {
        if ("".equals(updaterName)) updaterName = null;
        this.updaterName = updaterName;
    }

    public Date getCreatedAfter() {
        return createdAfter;
    }

    public void setCreatedAfter(Date createdAfter) {
        this.createdAfter = createdAfter;
    }

    public Date getCreatedBefore() {
        return createdBefore;
    }

    public void setCreatedBefore(Date createdBefore) {
        this.createdBefore = createdBefore;
    }

    public Date getUpdatedAfter() {
        return updatedAfter;
    }

    public void setUpdatedAfter(Date updatedAfter) {
        this.updatedAfter = updatedAfter;
    }

    public Date getUpdatedBefore() {
        return updatedBefore;
    }

    public void setUpdatedBefore(Date updatedBefore) {
        this.updatedBefore = updatedBefore;
    }

    public String getTags() {
        return tags;
    }

    public void setTags(String tags) {
        if ("".equals(tags)) {
            this.tags = null;
        } else {
            this.tags = tags.trim();
        }
    }

    public void setPropertyName(String propertyName) {
        if ("".equals(propertyName)) propertyName = null;
        this.propertyName = propertyName;
    }

    public void setPropertyValue(String propertyValue) {
        if ("".equals(propertyValue)) propertyValue = null;
        this.propertyValue = propertyValue;
    }

    public String getCommentWords() {
        return commentWords;
    }

    public void setCommentWords(String commentWords) {
        if ("".equals(commentWords)) commentWords = null;
        this.commentWords = commentWords;
    }

    private boolean queryExists(String queryPath) throws RegistryException {

        UserRegistry systemRegistry =
                (UserRegistry) System.getProperties().get(RegistryConstants.SYSTEM_REGISTRY);

        return systemRegistry.resourceExists(queryPath);
    }

    private void defineQuery(String queryPath) throws RegistryException {
        UserRegistry systemRegistry =
                (UserRegistry) System.getProperties().get(RegistryConstants.SYSTEM_REGISTRY);

        String queryContent = generateSQL();

        Resource q1 = systemRegistry.newResource();
        q1.setContent(queryContent);
        q1.setMediaType(RegistryConstants.SQL_QUERY_MEDIA_TYPE);
        q1.addProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME,
                RegistryConstants.RESOURCES_RESULT_TYPE);
        systemRegistry.put(queryPath, q1);
    }

    private String computeQueryPath() {

        if (queryPath == null) {

            StringBuffer buf = new StringBuffer("/system/queries/advanced");

            buf.append(resourceName != null ? "1" : "0");
            buf.append(authorName != null ? "1" : "0");
            buf.append(updaterName != null ? "1" : "0");
            buf.append(createdAfter != null ? "1" : "0");
            buf.append(createdBefore != null ? "1" : "0");
            buf.append(updatedAfter != null ? "1" : "0");
            buf.append(updatedBefore != null ? "1" : "0");
            buf.append(commentWords != null ? "1" : "0");
            buf.append(propertyName != null ? "1" : "0");
            buf.append(propertyValue != null ? "1" : "0");

            if (tags != null) {
                buf.append("T");
                buf.append(tags.split(" ").length);
            }

            queryPath = buf.toString();
        }

        return queryPath;
    }

    private String generateSQL() {
        ArrayList<String> tables = new ArrayList<String>();
        ArrayList<String> conditions = new ArrayList<String>();

        if (resourceName != null && resourceName.length() != 0) {
            conditions.add("R.PATH LIKE ?");
            params.add("%/" + resourceName);
        }

        if (authorName != null && authorName.length() != 0) {
            conditions.add("R.CREATOR LIKE ?");
            params.add(authorName);
        }

        if (updaterName != null && updaterName.length() != 0) {
            conditions.add("R.LAST_UPDATOR LIKE ?");
            params.add(updaterName);
        }

        if (createdAfter != null) {
            conditions.add("R.CREATED_TIME > ?");
            params.add(new Timestamp(createdAfter.getTime()).toString());
        }

        if (createdBefore != null) {
            conditions.add("R.CREATED_TIME < ?");
            params.add(new Timestamp(createdBefore.getTime()).toString());
        }

        if (updatedAfter != null) {
            conditions.add("R.LAST_UPDATED_TIME > ?");
            params.add(new Timestamp(updatedAfter.getTime()).toString());
        }

        if (updatedBefore != null) {
            conditions.add("R.LAST_UPDATED_TIME < ?");
            params.add(new Timestamp(updatedBefore.getTime()).toString());
        }

        if (commentWords != null && commentWords.length() != 0) {
            tables.add(", COMMENT C");
            conditions.add("C.RID=R.RID AND C.COMMENT_TEXT LIKE ?");
            params.add("%" + commentWords + "%");
        }

        if (tags != null && tags.length() > 0) {
            String[] tag = tags.split(" ");
            if (tag.length > 0) {
                tables.add(", TAG T");

                StringBuffer tagClause = new StringBuffer();
                tagClause.append("T.RID=R.RID AND (T.TAG_NAME=?");
                params.add(tag[0]);
                for (int i = 1; i < tag.length; i++) {
                    tagClause.append(" OR T.TAG_NAME=?");
                    params.add(tag[i]);
                }
                tagClause.append(")");
                conditions.add(tagClause.toString());
            }
        }

        if (propertyValue != null || propertyName != null) {
            tables.add(", PROPERTY P");
            StringBuffer propertyClause = new StringBuffer();
            propertyClause.append("P.RID=R.RID");
            if (propertyName != null) {
                propertyClause.append(" AND P.NAME=?");
                params.add(propertyName);                
            }
            if (propertyValue != null) {
                propertyClause.append(" AND P.VALUE LIKE ?");
                params.add("%" + propertyValue + "%");
            }
            conditions.add(propertyClause.toString());
        }

        StringBuffer query = new StringBuffer();
        query.append("SELECT R.PATH FROM RESOURCE R");
        for (String table : tables) {
            query.append(table);
        }
        boolean first = true;
        for (String condition : conditions) {
            if (first) {
                query.append(" WHERE ");
                first = false;
            } else {
                query.append(" AND ");
            }
            query.append(condition);
        }
        return query.toString();
    }
}
